# Python setup
In case a not installed python version is needed, follow these steps.

## Download Python
The required Python version can be downloaded from [here](https://www.python.org/downloads/).
Extract it and open a terminal in the extracted folder.

## Install Python
Configure the installation.
The `--with-openssl-rpath` is needed from Python 3.9 onwards, else pip will say it's missing ssl.
```sh
sudo apt install -y libssl-dev libffi-dev
./configure --enable-optimizations --with-openssl-rpath=/usr/include/x86_64-linux-gnu
```

Install Python
```sh
sudo make -j4 && sudo make altinstall
```

## Test
If everything is set up correctly, the following command should work
```sh
pythonx.y -m ssl
```

# TensorFlow setup
detailed instructions can be found at [tensorflow installation guide](https://www.tensorflow.org/install/pip), but they do not work out of the box!

## Conda env
### Install conda if not present
```sh
conda_install_script="Miniconda3-latest-Linux-x86_64.sh"
curl https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh -o $conda_install_script
bash $conda_install_script
rm $conda_install_script
```

### Setup conda env
```sh
conda create -n cuda12.3_py3.12 python=3.12
```
- You should choose a meaningful name (i.e. cuda12.3_py3.10 for cuda 12.3 in a python 3.12 environment)
- python=x.y where x.y is the python version number

Then you can activate the conda environment
```sh
conda activate cuda12.3_py3.12
```

### Install conda packages
```sh
conda install nvidia/label/cuda-x.y.z::cuda-toolkit
```
The correct package versions can be found in the [tensorflow versions](https://www.tensorflow.org/install/source#gpu) and the conda package name at the [anaconda package index](https://anaconda.org/nvidia/cuda-toolkit).

Then you can exit the conda env
```sh
conda deactivate
```

## Virtual environment
I highly recommend using virtualenv instead conda for virtualenv, because in my eyes conda is very dangerous. Conda links the executables from the base environment and filesystem in a virtual environment symbolically. This means, if you upgrade i.e. pip, the pip on the filesystem is also upgraded, which, in many cases, is unwanted behavior.

### Install virtualenv
Therefore I recommend the usage of a real virtual environment. This can be achieved for example, but not exclusively, with virtualenv
```sh
sudo apt update && sudo apt install -y virtualenv
```

### Setup virtual env
Navigate to your project folder and execute the following command
```sh
virtualenv venv --python=python3.12
```
the python version here and in the conda env should match

### Install packages
Activate the virtual env
```sh
source venv/bin/activate
```

Install prerequisites
```sh
pip install --upgrade setuptools wheel
```

Then install tensorflow
```sh
pip install nvidia-pyindex
pip install tensorflow[and-cuda]==x.y.z
```

To see which tensorrt version is needed run
```sh
strace -e open,openat python -c "import tensorflow as tf" 2>&1 | grep "libnvinfer\|TF-TRT"
```

Now, [tensorrt](https://docs.nvidia.com/deeplearning/tensorrt/) is installed (installation might fail, but as long as the libnvinver.so.\* and libnvinver_plugin.so.\* are present, that's no big deal)
```sh
pip install --no-cache tensorrt==x.y.z
```

Unfortunately, tensorrt does not link the shared libraries, so we need to do
```sh
tensorrt_dir=$(pwd)/venv/lib/python$(python -c "import sys; print(f'{sys.version_info.major}.{sys.version_info.minor}')")/site-packages/tensorrt_libs
ln -s $tensorrt_dir/libnvinfer.so.8 $tensorrt_dir/libnvinfer.so.8.6.1
ln -s $tensorrt_dir/libnvinfer_plugin.so.8 $tensorrt_dir/libnvinfer_plugin.so.8.6.1
```
Be aware, that your paths and versions may differ from the given paths.
Your paths can be found via
```sh
find . -iname *libnvinfer*
```

### Modify activation script
In order for tensorflow to find the cuda dependencies, these must be given in the virtual environment.
For this add these lines at the bottom of the activation script (`venv/bin/activate`)
```sh
# tf env
conda_tf_env="/home/andri/miniconda3/envs/cuda12.3_py3.12"
project_dir="/home/andri/Projects/tf_agents_tutorial"
virtualenv_site_packages="$project_dir/venv/lib/python3.12/site-packages"

_OLD_LD_LIBRARY_PATH="$LD_LIBRARY_PATH"
export _OLD_LD_LIBRARY_PATH
LD_LIBRARY_PATH="$LD_LIBRARY_PATH:$conda_tf_env/lib/:$virtualenv_site_packages/nvidia/cudnn/lib:$virtualenv_site_packages/tensorrt_libs/"
export LD_LIBRARY_PATH

_OLD_XLA_FLAGS="$XLA_FLAGS"
export _OLD_XLA_FLAGS
XLA_FLAGS="--xla_gpu_cuda_data_dir=$conda_tf_env"
export XLA_FLAGS
```
Be aware, that your paths may differ from the given paths.

Additionally add the following lines to the `deactivate()` function in the activation script
```sh
# tf env
LD_LIBRARY_PATH="$_OLD_LD_LIBRARY_PATH"
export LD_LIBRARY_PATH
unset _OLD_LD_LIBRARY_PATH

XLA_FLAGS="$_OLD_XLA_FLAGS"
export XLA_FLAGS
unset _OLD_XLA_FLAGS
```

## Test
After a deactivation & reactivation of the virtual environment the following command should show your GPUs
```sh
python3 -c "import tensorflow as tf; print(tf.config.list_physical_devices('GPU'))"
```

## A note about conda
conda is brutally bad, so please don't be silly and refuse to use it wherever possible
